from .colors import NamedColors
from .palettes import ColorLoversPalette, palette_blocks, palette_rectangles

from .interpolation import RGBa_interpolation
from .interpolation import LABa_interpolation

matplotlib_colors = NamedColors()
palette = ColorLoversPalette()
